// Represents a single set of blocks to run in order, waiting for each group's completion before
// moving on to the next one.

using Sandbox.ModAPI.Ingame;
using System.Collections.Generic;
using System.Linq;

namespace IngameScript
{
    partial class Program
    {
        public class Sequencer
        {
            public bool Running { get; private set; }

            private IConsole _console;
            private string _name;

            private SortedDictionary<int, List<ISequenceable>> _sequence;

            public Sequencer(string name, IConsole console)
            {
                _name = name;
                _console = console;
                _sequence = new SortedDictionary<int, List<ISequenceable>>();
            }

            public void AddBlock(ISequenceable block)
            {
                if (!_sequence.ContainsKey(block.Step)) _sequence[block.Step] = new List<ISequenceable>();
                _sequence[block.Step].Add(block);
            }

            // To activate the Sequencer, call this once, then call `MoveNext()` once per tick
            //   on the returned object until it returns false.
            //   (then be sure to call `Dispose() on that object`)
            public IEnumerator<bool> RunSequence(bool reverse = false)
            {
                if (Running)
                {
                    _console.Print("Already running, ignoring invocation.");
                    yield break;
                }

                // This is a bit convoluted, but we're extracting the iterator for use in the foreach directly.
                // This allows us to iterate in reverse when reversing/"closing" the sequence.
                IEnumerable<KeyValuePair<int, List<ISequenceable>>> steps = reverse ? _sequence.Reverse() : _sequence;
                foreach (KeyValuePair<int, List<ISequenceable>> kvp in steps)
                {
                    List<ISequenceable> blocks = kvp.Value;
                    Running = true;

                    foreach (ISequenceable block in blocks)
                    {
                        // TODO: add some sort of handling for block.Running == true
                        block.Start(reverse);
                    }
                    yield return true;

                    while (Running)
                    {
                        Running = false;
                        foreach (ISequenceable block in blocks)
                        {
                            if (block.Check()) Running = true;
                        }
                        yield return true;
                    }

                    foreach (ISequenceable block in blocks)
                    {
                        block.Finish();
                    }
                }

                yield break;
            }
        }
    }
}